@@ -9,9 +9,9 @@
 /***************************************************************************
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
- *   it under the terms of the GNU General Public License as published by  *
- *   the Free Software Foundation; either version 2 of the License, or     *
- *   (at your option) any later version.                                   *
+ *   it under the terms of the GNU Lesser General Public License as        *
+ *   published by the Free Software Foundation; either version 2 of the    *
+ *   License, or (at your option) any later version.                       *
  *                                                                         *
  ***************************************************************************/
 
@@ -205,6 +205,21 @@ namespace Qyoto {
 		[DllImport("libqyoto", CharSet=CharSet.Ansi)]
 		public static extern void InstallAddUIntToListUInt(AddUInt callback);
 		
+		[DllImport("libqyoto", CharSet=CharSet.Ansi)]
+		public static extern void InstallQPairGetFirst(GetIntPtr callback);
+
+		[DllImport("libqyoto", CharSet=CharSet.Ansi)]
+		public static extern void InstallQPairGetSecond(GetIntPtr callback);
+
+		[DllImport("libqyoto", CharSet=CharSet.Ansi)]
+		public static extern void InstallCreateQPair(CreateQPairFn callback);
+
+		[DllImport("libqyoto", CharSet=CharSet.Ansi)]
+		public static extern void InstallUnboxToStackItem(SetIntPtr callback);
+
+		[DllImport("libqyoto", CharSet=CharSet.Ansi)]
+		public static extern void InstallBoxFromStackItem(CreateInstanceFn callback);
+
 		[DllImport("libqyoto", CharSet=CharSet.Ansi)]
 		public static extern void InstallGenericPointerGetIntPtr(GetIntPtr callback);
 		
@@ -238,6 +253,7 @@ namespace Qyoto {
 		public delegate IntPtr ConstructDict(string type1, string type2);
 		public delegate void SetPropertyFn(IntPtr obj, string name, IntPtr variant);
 		public delegate void InvokeDelegateFn(Delegate d, IntPtr stack);
+		public delegate IntPtr CreateQPairFn(IntPtr first, IntPtr second);
 #endregion
 		
 #region marshalling functions
@@ -264,8 +280,12 @@ namespace Qyoto {
 			Object instance = ((GCHandle) instancePtr).Target;
 //			Debug.Assert(instance != null);
 
-			SmokeClassData data = GetSmokeClassData(instance.GetType());
-			return (IntPtr) data.smokeObjectField.GetValue(instance);
+			try {
+				SmokeClassData data = GetSmokeClassData(instance.GetType());
+				return (IntPtr) data.smokeObjectField.GetValue(instance);
+			} catch {
+				return IntPtr.Zero;
+			}
 		}
 		
 		public static void SetSmokeObject(IntPtr instancePtr, IntPtr smokeObjectPtr) {
@@ -363,7 +383,7 @@ namespace Qyoto {
 		public static void MapPointer(IntPtr ptr, IntPtr instancePtr, bool createGlobalReference) {
 			lock (pointerMap) {
 				Object instance = ((GCHandle) instancePtr).Target;
-				WeakReference weakRef = new WeakReference(instance);
+				WeakReference weakRef = new WeakReference(instance, true);
 				pointerMap[ptr] = weakRef;
 #if DEBUG
 				if ((QDebug.DebugChannel() & QtDebugChannel.QTDB_GC) != 0) {
@@ -411,45 +431,79 @@ namespace Qyoto {
 		// of a Qyoto class and therefore could have custom slots or overriden methods
 		public static IntPtr GetInstance(IntPtr ptr, bool allInstances) {
 			WeakReference weakRef;
-			if (!pointerMap.TryGetValue(ptr, out weakRef)) {
+			object strongRef;
+			lock (pointerMap) {
+				if (!pointerMap.TryGetValue(ptr, out weakRef)) {
 #if DEBUG
-				if (	(QDebug.DebugChannel() & QtDebugChannel.QTDB_GC) != 0
-						&& QDebug.debugLevel >= DebugLevel.Extensive ) 
-				{
-					Console.WriteLine("GetInstance() pointerMap[0x{0:x8}] == null", (int) ptr);
-				}
+					if (	(QDebug.DebugChannel() & QtDebugChannel.QTDB_GC) != 0
+							&& QDebug.debugLevel >= DebugLevel.Extensive ) 
+					{
+						Console.WriteLine("GetInstance() pointerMap[0x{0:x8}] == null", (int) ptr);
+					}
 #endif
-				return IntPtr.Zero;
-			}
-
-			if (weakRef.IsAlive) {
-#if DEBUG
-				if (	(QDebug.DebugChannel() & QtDebugChannel.QTDB_GC) != 0
-						&& QDebug.debugLevel >= DebugLevel.Extensive ) 
-				{
-					Console.WriteLine("GetInstance() weakRef.IsAlive 0x{0:x8} -> {1}", (int) ptr, weakRef.Target);
-				}
-#endif
-				if (!allInstances && IsSmokeClass(weakRef.Target.GetType())) {
 					return IntPtr.Zero;
-				} 
+				}
+
+				if (weakRef != null && weakRef.IsAlive) {
+#if DEBUG
+					if (	(QDebug.DebugChannel() & QtDebugChannel.QTDB_GC) != 0
+							&& QDebug.debugLevel >= DebugLevel.Extensive ) 
+					{
+						Console.WriteLine("GetInstance() weakRef.IsAlive 0x{0:x8} -> {1}", (int) ptr, weakRef.Target);
+					}
+#endif
+					if (!allInstances && IsSmokeClass(weakRef.Target.GetType())) {
+						return IntPtr.Zero;
+					} 
 
 #if DEBUG
-				GCHandle instanceHandle = DebugGCHandle.Alloc(weakRef.Target);
+					GCHandle instanceHandle = DebugGCHandle.Alloc(weakRef.Target);
 #else
-				GCHandle instanceHandle = GCHandle.Alloc(weakRef.Target);
+					GCHandle instanceHandle = GCHandle.Alloc(weakRef.Target);
 #endif
-				return (IntPtr) instanceHandle;
-			} else {
+					return (IntPtr) instanceHandle;
+				} else if (Environment.HasShutdownStarted && globalReferenceMap.TryGetValue(ptr, out strongRef)) {
 #if DEBUG
-				if (	(QDebug.DebugChannel() & QtDebugChannel.QTDB_GC) != 0
-						&& QDebug.debugLevel >= DebugLevel.Extensive ) 
-				{
-					Console.WriteLine("GetInstance() weakRef dead ptr: 0x{0:x8}", (int) ptr);
-				}
+					if (	(QDebug.DebugChannel() & QtDebugChannel.QTDB_GC) != 0
+							&& QDebug.debugLevel >= DebugLevel.Extensive ) 
+					{
+						Console.WriteLine("GetInstance() strongRef 0x{0:x8} -> {1}", (int) ptr, strongRef);
+					}
 #endif
-				pointerMap.Remove(ptr);
-				return IntPtr.Zero;
+					if (!allInstances && IsSmokeClass(strongRef.GetType())) {
+						return IntPtr.Zero;
+					} 
+
+#if DEBUG
+					GCHandle instanceHandle = DebugGCHandle.Alloc(strongRef);
+#else
+					GCHandle instanceHandle = GCHandle.Alloc(strongRef);
+#endif
+					return (IntPtr) instanceHandle;
+				} else {
+#if DEBUG
+					if (	(QDebug.DebugChannel() & QtDebugChannel.QTDB_GC) != 0
+							&& QDebug.debugLevel >= DebugLevel.Extensive ) 
+					{
+						Console.WriteLine("GetInstance() weakRef dead ptr: 0x{0:x8}", (int) ptr);
+					}
+#endif
+					pointerMap.Remove(ptr);
+					return IntPtr.Zero;
+				}
+			}
+		}
+
+		// converts weak references to strong references so they are still available
+		// when the application is shutting down.
+		public static void ConvertRefs() {
+			lock (pointerMap) {
+				foreach (KeyValuePair<IntPtr, WeakReference> pair in pointerMap) {
+					if (!pair.Value.IsAlive)
+						continue;
+					lock (globalReferenceMap)
+						globalReferenceMap[pair.Key] = pair.Value.Target;
+				}
 			}
 		}
 
@@ -525,8 +579,12 @@ namespace Qyoto {
 		// The class is not a custom subclass of a Qyoto class, and also is not
 		// a superclass of a Qyoto class, such as a MarshalByRefObject.
 		public static bool IsSmokeClass(Type klass) {
-			SmokeClassData data = GetSmokeClassData(klass);
-			return data != null && data.className != null;
+			try {
+				SmokeClassData data = GetSmokeClassData(klass);
+				return data != null && data.className != null;
+			} catch {
+				return false;
+			}
 		}
 
 		// The C++ class name signature of a Smoke class or interface
@@ -594,9 +652,9 @@ namespace Qyoto {
 
 		public static IntPtr IntPtrFromString(string str) {
 #if DEBUG
-			return (IntPtr) DebugGCHandle.Alloc(str);
+			return (IntPtr) DebugGCHandle.Alloc(string.Copy(str));
 #else
-			return (IntPtr) GCHandle.Alloc(str);
+			return (IntPtr) GCHandle.Alloc(string.Copy(str));
 #endif
 		}
 
@@ -828,6 +886,98 @@ namespace Qyoto {
 			return map;
 		}
 		
+		public static IntPtr QPairGetFirst(IntPtr pair) {
+			object o = ((GCHandle) pair).Target;
+			object field = o.GetType().GetField("first").GetValue(o);
+			if (field == null) return IntPtr.Zero;
+			return (IntPtr) GCHandle.Alloc(field);
+		}
+		
+		public static IntPtr QPairGetSecond(IntPtr pair) {
+			object o = ((GCHandle) pair).Target;
+			object field = o.GetType().GetField("second").GetValue(o);
+			if (field == null) return IntPtr.Zero;
+			return (IntPtr) GCHandle.Alloc(field);
+		}
+		
+		public static IntPtr CreateQPair(IntPtr first, IntPtr second) {
+			Type t = typeof(QPair<,>);
+			object firstObject = ((GCHandle) first).Target;
+			object secondObject = ((GCHandle) second).Target;
+			t = t.MakeGenericType(new Type[] { firstObject.GetType(), secondObject.GetType() });
+			object pair = Activator.CreateInstance(t, new object[] { firstObject, secondObject });
+			return (IntPtr) GCHandle.Alloc(pair);
+		}
+		
+		public static unsafe void UnboxToStackItem(IntPtr obj, IntPtr st) {
+			StackItem *item = (StackItem*) st;
+			object o = ((GCHandle) obj).Target;
+			Type t = o.GetType();
+			if (t == typeof(int) || t.IsEnum) {
+				item->s_int = (int) o;
+			} else if (t == typeof(bool)) {
+				item->s_bool = (bool) o;
+			} else if (t == typeof(short)) {
+				item->s_short = (short) o;
+			} else if (t == typeof(float)) {
+				item->s_float = (float) o;
+			} else if (t == typeof(double)) {
+				item->s_double = (double) o;
+			} else if (t == typeof(long)) {
+				item->s_long = (long) o;
+			} else if (t == typeof(ushort)) {
+				item->s_ushort = (ushort) o;
+			} else if (t == typeof(uint)) {
+				item->s_uint = (uint) o;
+			} else if (t == typeof(ulong)) {
+				item->s_ulong = (ulong) o;
+			} else if (t == typeof(sbyte)) {
+				item->s_char = (sbyte) o;
+			} else if (t == typeof(byte)) {
+				item->s_uchar = (byte) o;
+			} else if (t == typeof(char)) {
+				item->s_char = (sbyte) (char) o;
+			} else {
+				item->s_class = IntPtr.Zero;
+			}
+		}
+		
+		public static unsafe IntPtr BoxFromStackItem(string type, IntPtr st) {
+			Type t = Type.GetType(type);
+			if (t == null) return IntPtr.Zero;
+			StackItem *item = (StackItem*) st;
+			object box = null;
+			
+			if (t == typeof(int)) {
+				box = item->s_int;
+			} else if (t == typeof(bool)) {
+				box = item->s_bool;
+			} else if (t == typeof(short)) {
+				box = item->s_short;
+			} else if (t == typeof(float)) {
+				box = item->s_float;
+			} else if (t == typeof(double)) {
+				box = item->s_double;
+			} else if (t == typeof(long)) {
+				box = (SizeOfNativeLong < sizeof(long))? item->s_int : item->s_long;
+			} else if (t == typeof(ushort)) {
+				box = item->s_ushort;
+			} else if (t == typeof(uint)) {
+				box = item->s_uint;
+			} else if (t == typeof(ulong)) {
+				box = (SizeOfNativeLong < sizeof(long))? item->s_uint : item->s_ulong;
+			} else if (t == typeof(sbyte)) {
+				box = item->s_char;
+			} else if (t == typeof(byte)) {
+				box = item->s_uchar;
+			} else if (t == typeof(char)) {
+				box = (char) item->s_char;
+			}
+			
+			if (box == null) return IntPtr.Zero;
+			return (IntPtr) GCHandle.Alloc(box);
+		}
+		
 		public static IntPtr GenericPointerGetIntPtr(IntPtr obj) {
 			object o = ((GCHandle) obj).Target;
 			return (IntPtr) o.GetType().GetMethod("ToIntPtr").Invoke(o, null);
@@ -897,6 +1047,13 @@ namespace Qyoto {
 			InstallGetProperty(GetProperty);
 			InstallSetProperty(SetProperty);
 			
+			InstallQPairGetFirst(QPairGetFirst);
+			InstallQPairGetSecond(QPairGetSecond);
+			InstallCreateQPair(CreateQPair);
+			
+			InstallUnboxToStackItem(UnboxToStackItem);
+			InstallBoxFromStackItem(BoxFromStackItem);
+			
 			InstallGenericPointerGetIntPtr(GenericPointerGetIntPtr);
 			InstallCreateGenericPointer(CreateGenericPointer);
 			
